@@ -66,8 +66,9 @@ qpwgraph_canvas::qpwgraph_canvas ( QWidget *parent )
 		m_connect(nullptr), m_rubberband(nullptr),
 		m_zoom(1.0), m_zoomrange(false),
 		m_commands(nullptr), m_settings(nullptr), m_patchbay(nullptr),
-		m_selected_nodes(0), m_edit_item(nullptr),
-		m_editor(nullptr), m_edited(0)
+		m_patchbay_edit(false), m_patchbay_autopin(true),
+		m_selected_nodes(0), m_repel_overlapping_nodes(false),
+		m_edit_item(nullptr), m_editor(nullptr), m_edited(0)
 {
 	m_scene = new QGraphicsScene();
 
@@ -143,6 +144,119 @@ qpwgraph_patchbay *qpwgraph_canvas::patchbay (void) const
 }
 
 
+// Patchbay auto-pin accessors.
+void qpwgraph_canvas::setPatchbayAutoPin ( bool on )
+{
+	m_patchbay_autopin = on;
+}
+
+
+bool qpwgraph_canvas::isPatchbayAutoPin (void) const
+{
+	return m_patchbay_autopin;
+}
+
+
+// Patchbay edit-mode accessors.
+void qpwgraph_canvas::setPatchbayEdit ( bool on )
+{
+	if (m_patchbay == nullptr)
+		return;
+
+	if ((!on && !m_patchbay_edit) ||
+		( on &&  m_patchbay_edit))
+		return;
+
+	m_patchbay_edit = on;
+
+	patchbayEdit();
+}
+
+
+bool qpwgraph_canvas::isPatchbayEdit (void) const
+{
+	return (m_patchbay && m_patchbay_edit);
+}
+
+
+void qpwgraph_canvas::patchbayEdit (void)
+{
+	foreach (QGraphicsItem *item, m_scene->items()) {
+		if (item->type() == qpwgraph_connect::Type) {
+			qpwgraph_connect *connect = static_cast<qpwgraph_connect *> (item);
+			if (connect) {
+				connect->setDimmed(
+					m_patchbay_edit && !m_patchbay->findConnect(connect));
+			}
+		}
+	}
+}
+
+
+bool qpwgraph_canvas::canPatchbayPin (void) const
+{
+	if (m_patchbay == nullptr || !m_patchbay_edit)
+		return false;
+
+	foreach (QGraphicsItem *item, m_scene->selectedItems()) {
+		if (item->type() == qpwgraph_connect::Type) {
+			qpwgraph_connect *connect = static_cast<qpwgraph_connect *> (item);
+			if (connect && !m_patchbay->findConnect(connect))
+				return true;
+		}
+	}
+
+	return false;
+}
+
+
+bool qpwgraph_canvas::canPatchbayUnpin (void) const
+{
+	if (m_patchbay == nullptr || !m_patchbay_edit)
+		return false;
+
+	foreach (QGraphicsItem *item, m_scene->selectedItems()) {
+		if (item->type() == qpwgraph_connect::Type) {
+			qpwgraph_connect *connect = static_cast<qpwgraph_connect *> (item);
+			if (connect && m_patchbay->findConnect(connect))
+				return true;
+		}
+	}
+
+	return false;
+}
+
+
+void qpwgraph_canvas::patchbayPin (void)
+{
+	if (m_patchbay == nullptr || !m_patchbay_edit)
+		return;
+
+	foreach (QGraphicsItem *item, m_scene->selectedItems()) {
+		if (item->type() == qpwgraph_connect::Type) {
+			qpwgraph_connect *connect = static_cast<qpwgraph_connect *> (item);
+			if (connect && m_patchbay->connect(connect, true))
+				connect->setDimmed(false);
+		}
+	}
+}
+
+
+void qpwgraph_canvas::patchbayUnpin (void)
+{
+	if (m_patchbay == nullptr || !m_patchbay_edit)
+		return;
+
+	foreach (QGraphicsItem *item, m_scene->selectedItems()) {
+		if (item->type() == qpwgraph_connect::Type) {
+			qpwgraph_connect *connect = static_cast<qpwgraph_connect *> (item);
+			if (connect && m_patchbay->connect(connect, false))
+				connect->setDimmed(true);
+		}
+	}
+}
+
+
 // Canvas methods.
 void qpwgraph_canvas::addItem ( qpwgraph_item *item )
 {
@@ -155,7 +269,9 @@ void qpwgraph_canvas::addItem ( qpwgraph_item *item )
 			m_nodes.append(node);
 			m_node_ids.insert(qpwgraph_node::NodeIdKey(node), node);
 			m_node_keys.insert(qpwgraph_node::NodeNameKey(node), node);
-			if (!restoreNode(node))
+			if (restoreNode(node))
+				emit updated(node);
+			else
 				emit added(node);
 		}
 	}
@@ -165,6 +281,14 @@ void qpwgraph_canvas::addItem ( qpwgraph_item *item )
 		if (port)
 			restorePort(port);
 	}
+	else
+	if (item->type() == qpwgraph_connect::Type) {
+		qpwgraph_connect *connect = static_cast<qpwgraph_connect *> (item);
+		if (connect) {
+			connect->setDimmed(m_patchbay_edit &&
+				m_patchbay && !m_patchbay->findConnect(connect));
+		}
+	}
 }
 
 
@@ -393,7 +517,7 @@ QList<qpwgraph_node *> qpwgraph_canvas::findNodes (
 void qpwgraph_canvas::emitConnectPorts (
 	qpwgraph_port *port1, qpwgraph_port *port2, bool is_connect )
 {
-	if (m_patchbay)
+	if (m_patchbay && (m_patchbay_autopin || !is_connect))
 		m_patchbay->connectPorts(port1, port2, is_connect);
 
 	if (is_connect)
@@ -1451,4 +1575,95 @@ void qpwgraph_canvas::editingFinished (void)
 }
 
 
+// Repel overlapping nodes...
+void qpwgraph_canvas::setRepelOverlappingNodes ( bool on )
+{
+	m_repel_overlapping_nodes = on;
+}
+
+
+bool qpwgraph_canvas::isRepelOverlappingNodes (void) const
+{
+	return m_repel_overlapping_nodes;
+}
+
+
+void qpwgraph_canvas::repelOverlappingNodes ( qpwgraph_node *node,
+	qpwgraph_move_command *move_command, const QPointF& delta )
+{
+	const qreal MIN_NODE_GAP = 8.0f;
+
+	node->setMarked(true);
+
+	QRectF rect1 = node->sceneBoundingRect();
+	rect1.adjust(
+		-2.0 * MIN_NODE_GAP, -MIN_NODE_GAP,
+		+2.0 * MIN_NODE_GAP, +MIN_NODE_GAP);
+
+	foreach (qpwgraph_node *node2, m_nodes) {
+		if (node2->isMarked())
+			continue;
+		const QPointF& pos1
+			= node2->pos();
+		QPointF pos2 = pos1;
+		const QRectF& rect2
+			= node2->sceneBoundingRect();
+		const QRectF& recti
+			= rect2.intersected(rect1);
+		if (!recti.isNull()) {
+			const QPointF delta2
+				= (delta.isNull() ? rect2.center() - rect1.center() : delta);
+			if (recti.width() < (1.5 * recti.height())) {
+				qreal dx = recti.width();
+				if ((delta2.x() < 0.0 && recti.width() >= rect1.width()) ||
+					(delta2.x() > 0.0 && recti.width() >= rect2.width())) {
+					dx += qAbs(rect2.right() - rect1.right());
+				}
+				else
+				if ((delta2.x() > 0.0 && recti.width() >= rect1.width()) ||
+					(delta2.x() < 0.0 && recti.width() >= rect2.width())) {
+					dx += qAbs(rect2.left() - rect1.left());
+				}
+				if (delta2.x() < 0.0)
+					pos2.setX(pos1.x() - dx);
+				else
+					pos2.setX(pos1.x() + dx);
+			} else {
+				qreal dy = recti.height();
+				if ((delta2.y() < 0.0 && recti.height() >= rect1.height()) ||
+					(delta2.y() > 0.0 && recti.height() >= rect2.height())) {
+					dy += qAbs(rect2.bottom() - rect1.bottom());
+				}
+				else
+				if ((delta2.y() > 0.0 && recti.height() >= rect1.height()) ||
+					(delta2.y() < 0.0 && recti.height() >= rect2.height())) {
+					dy += qAbs(rect2.top() - rect1.top());
+				}
+				if (delta2.y() < 0.0)
+					pos2.setY(pos1.y() - dy);
+				else
+					pos2.setY(pos1.y() + dy);
+			}
+			// Repel this node...
+			node2->setPos(pos2);
+			// Add this node for undo/redo...
+			if (move_command)
+				move_command->addItem(node2, pos1, pos2);
+			// Repel this node neighbors, if any...
+			repelOverlappingNodes(node2, move_command, delta2);
+		}
+	}
+
+	node->setMarked(false);
+}
+
+
+void qpwgraph_canvas::repelOverlappingNodesAll (
+	qpwgraph_move_command *move_command )
+{
+	foreach (qpwgraph_node *node, m_nodes)
+		repelOverlappingNodes(node, move_command);
+}
+
+
 // end of qpwgraph_canvas.cpp
